/***************************************************************************
   network_icmp.cpp  -  ICMP network plugin
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: network_icmp.c,v 1.5 2002/08/22 21:54:37 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file network_icmp.c
  * File contains ICMP plugin functions
  * This file is a good example of the plugin API.
  * Use it as an example for your own plugin development.
  */

#include <assert.h>
#include "network_icmp.h"

static const char rcsid[] =
        "\100$ autosec/sonar: $Id: network_icmp.c,v 1.5 2002/08/22 21:54:37 red0x Exp $";

static char *cmd;

/** @name Plugin API -- Plugin side
  * Every plugin you make must have a plugin_init function
  * exported as a C function, so its name wont get mangled.
  * The duty of this function is to fill a plugin_t structure
  * with function pointers to your plugin's functions and do
  * any other setup, and then return the address of the plugin_t
  * structure.
  * That is the only Mandatorily named function in your plugin.
 */
//@{
/** @fn plugin_init(void *in_data)
  * Plugin Init function, part of the Plugin API
  * @param in_data Pointer to a sonar_t structure that will be 
  *  passed from sonar when your plugin is loaded.
  * @return A pointer to a filled plugin_t structure.
  */
void *plugin_init(void *in_data)
{
        assert(in_data != NULL);
        //useless reference, to avoid a warning
        if(rcsid)
        {
        }
        // this is assumed
        mthis = (sonar_t *) in_data;

        srand(getpid() ^ (uid_t) time(NULL));
        my_errno = PLUGIN_LOADED;
        my_socket = -1;

        //set some defaults
        ttl = 255;
        delay = -1;
        seq_no = (int) (256.0 * rand() / (RAND_MAX + 0.0));
        size = 0;
        payload_file = NULL;

        //setup my plugin stuff
        pthis.type = NETWORK_PLUGIN;
        pthis.errno = &my_errno;
        pthis.ready = false;
        pthis.plugin_init = &plugin_init;
        pthis.plugin_destroy = &icmp_end;
        pthis.plugin_open = &icmp_leadin;
        pthis.plugin_close = &icmp_close;
        pthis.plugin_next = &icmp_add;
        pthis.plugin_sweep = &icmp_sweep;
        pthis.plugin_args = &icmp_args;
        pthis.plugin_sweep = &icmp_sweep;
        // red0x: do we even need this??
        pthis.plugin_go = NULL;
        pthis.plugin_usage = &icmp_usage;
        pthis.plugin_status = &icmp_status;
        pthis.plugin_last_error = &icmp_last_error;
        ready = false;
        hosts = (host_list_t *) malloc(sizeof(host_list_t));
        hosts->next = NULL;
        hosts->host = NULL;
        cmd = NULL;
        hosts->ip = 0;
        first_host = hosts;
        num_hosts = 0;
	my_type = ICMP_ECHO;
	my_code = 0;

        return &pthis;
}

/** @fn icmp_end(void)
  * Plugin Unload function
  * @return PLUGIN_UNLOADED
  * @sideeffect Plugin will be unloaded, and should be unuable
  */
static int icmp_end(void)
{
        my_errno = PLUGIN_UNLOADED;
        if(my_socket != -1 && my_socket != 0)
        {
                close(my_socket);
                my_socket = -1;
        }

        ttl = 0;
        delay = 0;
        seq_no = 0;
        size = 0;
        if(payload_file != NULL)
        {
                free(payload_file);
                payload_file = NULL;
        }

        pthis.type = NETWORK_PLUGIN;
        pthis.errno = &my_errno;
        pthis.ready = false;
        pthis.plugin_init = NULL;
        pthis.plugin_destroy = NULL;
        pthis.plugin_open = NULL;
        pthis.plugin_close = NULL;
        pthis.plugin_next = NULL;
        pthis.plugin_sweep = NULL;
        pthis.plugin_args = NULL;
        pthis.plugin_sweep = NULL;
        // red0x: do we even need this??
        pthis.plugin_go = NULL;
        pthis.plugin_usage = NULL;
        pthis.plugin_status = NULL;
        pthis.plugin_last_error = NULL;
        ready = false;
        assert(hosts != NULL);
        free(hosts);
        hosts = NULL;
        if(cmd != NULL)
        {
                free(cmd);
                cmd = NULL;
        }

        num_hosts = 0;
        mthis = NULL;
        first_host = NULL;
        return PLUGIN_UNLOADED;
}

/** @fn icmp_add(char *hostname)
  * Plugin Next function
  * @param hostname Hostname to add to queue
  * @return PLUGIN_OK if host was added, PLUGIN_ERROR otherwise
  * @sideeffect Tries to resolve hostname
  */
static int icmp_add(char *host_name)
{
        char t;
        char *range, *tmp;
        char *hostname;
        {
                assert(mthis != NULL);
                assert(mthis->memput != NULL);
                assert(hosts != NULL);
                assert(mthis->resolver != NULL);
        }

        t = (char) *host_name;
        if(t != 'I')
                return PLUGIN_ERROR;
        hostname = host_name + sizeof(char);

        if((range = strchr((const char *) hostname, '[')) != NULL)
        {
                if((tmp = strchr((const char *) hostname, ']')) != NULL)
                {
                        //we have a valid range of the form "[1-255]" somewhere
                        // what we're gunna do is solve the first range, and let
                        // the recursion unroll the other ranges
                        //we need the character after the first [
                        char *host, *head, *zero, *tail, *h2, t;
                        int start, end, j, len;

                        // leave the 'A' prefix for now
                        hostname = host_name;

                        // save the precursor to the range                     
                        t = range[0];
                        range[0] = 0x00;
                        head = (char *) malloc(strlen(hostname));
                        strncpy(head, hostname, strlen(hostname));
                        range[0] = t;

                        // get the bound of the range
                        zero = strchr((const char *) range, '-');
                        if(zero == NULL)
                                mthis->usage("invalid range");
                        *zero = 0x00;
                        start = atoi((const char *) (range + sizeof(char)));
                        end = atoi((const char *) (zero + sizeof(char)));

                        // save everything after the range
                        len = strlen(tmp);
                        tail = (char *) malloc(len);
                        memset(tail, 0, len);
                        strncpy(tail, (const char *) (tmp + sizeof(char)),
                                len - 1);

                        // now we replace the '[%d-%d]' with a %d
                        host = (char *) malloc(strlen(head) + 3 +
                                               strlen(tail) + 5);
                        strncpy(host, head, strlen(head));
                        strncat(host, "%d", 2 * sizeof(char));
                        strncat(host, tail, strlen(tail));

                        for(j = start; j <= end; j++)
                        {
                                h2 = (char *) malloc(strlen(host) + 5);
                                memset(h2, 0, strlen(host) + 5);
                                snprintf(h2, strlen(host) + 5, host, j);

                                if(icmp_add(h2) != PLUGIN_OK)
                                        mthis->usage("invalid range");

                        }
                        return PLUGIN_OK;
                }
                else
                {
                        mthis->usage("invalid range");
                        //invalid range
                }
        }
        if((range = strchr((const char *) hostname, '*')) != NULL)
        {
                //we have a * range like 192.168.1.*
                int start = 1, end = 255, j, tlen, len;
                char *head, *tail, t, *host, *h2;

                // leave the 'A' intact
                hostname = host_name;

                // save the head
                t = range[0];
                range[0] = 0x00;
                head = (char *) malloc(strlen(hostname));
                strncpy(head, hostname, strlen(hostname));
                range[0] = t;

                //save the tail
                range++;
                tlen = strlen(range);
                tail = (char *) malloc(tlen + 1);
                memset(tail, 0, tlen + 1);
                strncpy(tail, range, tlen);

                //build the format string
                len = strlen(head) + strlen(tail) + (sizeof(char) * 2) + 5;
                host = (char *) malloc(len);
                memset(host, 0, len);
                strncpy(host, head, strlen(head));
                strncat(host, "%d", sizeof(char) * 2);
                strncat(host, tail, strlen(tail));
                for(j = start; j <= end; j++)
                {
                        h2 = (char *) malloc(strlen(host) + 5);
                        memset(h2, 0, strlen(host) + 5);
                        snprintf(h2, strlen(host) + 5, host, j);

                        if(icmp_add(h2) != PLUGIN_OK)
                                mthis->usage("invalid range");
                }

                return PLUGIN_OK;
        }
        if((range = strchr((const char *) hostname, '/')) != NULL)
        {
                //we have a /24 or /16 type range
                int a, b, c, d, mask, len;
                char *host, *h2;
                range[0] = 0x00;
                range++;
                mask = atoi(range);

                if(mthis->ipverify(hostname) == 0)
                {
                        //we have a hostname
                        unsigned int ip;
                        struct in_addr tmp;
                        struct hostent *he;
                        struct sa tp;

                        if(mthis->ipverify(hostname))
                        {
                                ip = inet_addr(hostname);
                        }
                        else
                        {
                                he = gethostbyname(hostname);
                                if(he)
                                {
                                        memcpy(&tp.addr, he->h_addr,
                                               he->h_length);
                                        ip = tp.addr;
                                        he = gethostbyaddr(he->h_addr, 4,
                                                           AF_INET);
                                        if(!he)
                                                ip = mthis->
                                                        resolver(hostname);
                                }
                                else
                                        return PLUGIN_ERROR;
                        }

                        if(ip == INADDR_NONE)
                        {
                                ip = 0;
                                return PLUGIN_ERROR;
                        }

                        if(ip == INADDR_NONE)
                                mthis->usage("unable to resolve hostname");
                        tmp.s_addr = ip;
                        hostname = inet_ntoa(tmp);

                }
                switch (mask)
                {
                        case 0:
                                mthis->memput(&host, "I*.*.*.*\0",
                                              strlen("A*.*.*.*\0"));
                                if(icmp_add(host) != PLUGIN_OK)
                                        mthis->usage
                                                ("failure scanning whole internet");
                                return PLUGIN_OK;
                                break;
                        case 8:
                                mthis->memput(&host, "I%d.%d.%d.%d\0",
                                              strlen("A%d.%d.%d.%d\0"));
                                len = strlen(hostname) + 5;
                                h2 = (char *) malloc(len);
                                memset(h2, 0, len);
                                sscanf(hostname, "%d", &a);
                                for(b = 1; b <= 255; b++)
                                {
                                        for(c = 1; c <= 255; c++)
                                        {
                                                for(d = 1; d <= 255; d++)
                                                {
                                                        snprintf(h2, len,
                                                                 host, a, b,
                                                                 c, d);
                                                        if(icmp_add(h2) !=
                                                           PLUGIN_OK)
                                                                mthis->usage
                                                                        ("failure scanning class A");
                                                }
                                        }
                                }
                                return PLUGIN_OK;
                                break;
                        case 16:
                                mthis->memput(&host, "I%d.%d.%d.%d\0",
                                              strlen("A%d.%d.%d.%d\0"));
                                len = strlen(hostname) + 5;
                                h2 = (char *) malloc(len);
                                memset(h2, 0, len);
                                sscanf(hostname, "%d.%d", &a, &b);
                                for(c = 1; c <= 255; c++)
                                {
                                        for(d = 1; d <= 255; d++)
                                        {
                                                snprintf(h2, len, host, a, b,
                                                         c, d);
                                                if(icmp_add(h2) != PLUGIN_OK)
                                                        mthis->usage
                                                                ("failure scanning class B");
                                        }
                                }
                                return PLUGIN_OK;
                                break;
                        case 24:
                                mthis->memput(&host, "I%d.%d.%d.%d\0",
                                              strlen("A%d.%d.%d.%d\0"));
                                len = strlen(hostname) + 5;
                                h2 = (char *) malloc(len);
                                memset(h2, 0, len);
                                sscanf(hostname, "%d.%d.%d", &a, &b, &c);
                                for(d = 1; d <= 255; d++)
                                {
                                        snprintf(h2, len, host, a, b, c, d);
                                        if(icmp_add(h2) != PLUGIN_OK)
                                                mthis->usage
                                                        ("failure scanning class C");
                                }
                                return PLUGIN_OK;
                                break;
                        case 32:
                                range[0] = 0x00;
                                if(icmp_add(host_name) != PLUGIN_OK)
                                        mthis->usage("Invalid hostname/ip");
                                return PLUGIN_OK;
                                break;
                        default:
                                mthis->usage
                                        ("masks must be either /0 (whole internet), /8 (class A)"
                                         " /16 (class B), /24 (class C), or /32 (single host)");
                }

        }
        else
        {
                struct hostent *he;
                struct sa tp;

                mthis->memput(&(hosts->host), hostname, strlen(hostname));

                if(mthis->ipverify(hostname))
                {
                        hosts->ip = inet_addr(hostname);
                        mthis->memput(&(hosts->host), hostname,
                                      strlen(hostname));
                }
                else
                {
                        he = gethostbyname(hostname);
                        if(he)
                        {
                                memcpy(&tp.addr, he->h_addr, he->h_length);
                                hosts->ip = tp.addr;
                                he = gethostbyaddr(he->h_addr, 4, AF_INET);
                                if(he)
                                        mthis->memput(&(hosts->host),
                                                      he->h_name,
                                                      strlen(he->h_name));
                                else
                                        mthis->memput(&(hosts->host),
                                                      hostname,
                                                      strlen(hostname));
                        }
                        else
                                return PLUGIN_ERROR;
                }

                if(hosts->ip == INADDR_NONE)
                {
                        free(hosts->host);
                        hosts->ip = 0;
                        return PLUGIN_ERROR;
                }

                hosts->next = (host_list_t *) malloc(sizeof(host_list_t));
                hosts = hosts->next;
                hosts->next = NULL;
                if(!ready || !pthis.ready)
                        icmp_leadin(NULL, 0);
                num_hosts++;
                //we have a host, so we are ready.
                ready = true;
                pthis.ready = true;
        }
        return PLUGIN_OK;
}

/** @fn icmp_args(char opt, char *optarg)
  * Plugin Argument handler
  * @param opt Option char as passed from getopt
  * @param optarg Option argument as passed from getopt
  * @return PLUGIN_OK if that argument was handled, PLUGIN_ERROR otherwise
  */
static int icmp_args(char opt, char *optarg)
{
	int mlen;
        {
                assert(mthis != NULL);
                assert(mthis->memput != NULL);
        }

        switch (opt)
        {
                case 'd':
                        delay = atoi(optarg);
                        if(delay < 0)
                                delay = 0;
                        break;
                case 'p':
                        mthis->memput(&payload_file, optarg, strlen(optarg));
                        break;
                case 's':
                        size = atoi(optarg);
                        if(size > 350)
                                size = 350;
                        break;
                case 't':
                        ttl = atoi(optarg);
                        break;
                case 'e':
                        mthis->memput(&cmd, optarg, strlen(optarg));
                        icmp_escape_shell(cmd);
                        break;
	        case 'i':
			mlen = strlen(optarg);
			my_type = -1;
			if(strncmp(optarg, "echo", (mlen < 4) ? mlen : 4) == 0)

				my_type = ICMP_ECHO;
			else if(strncmp(optarg, "du", (mlen < 2) ? mlen : 2) == 0)
				my_type = ICMP_UNREACH;
			else if(strncmp(optarg, "mask", (mlen < 4) ? mlen : 4) == 0)		
				my_type = ICMP_ADDRESS;
			else if(strncmp(optarg, "red", (mlen < 3) ? mlen : 3) == 0)
				my_type = ICMP_REDIRECT;
			else if(strncmp(optarg, "ts", (mlen < 2) ? mlen : 2) == 0)
				my_type = ICMP_TSTAMP;
			else if(strncmp(optarg, "info", (mlen < 4) ? mlen : 4) == 0)
				my_type = ICMP_IREQ;
			else if(strncmp(optarg, "param", (mlen < 5) ? mlen : 5) == 0)
				my_type = ICMP_PARAMPROB;
			else if(strncmp(optarg, "help", (mlen < 4) ? mlen : 4) == 0)
				mthis->usage("Currently supported types are:\n"
					     "\t echo - echo request\n"
					     "\t du - dest unreachable\n"
					     "\t mask - mask request\n"
					     "\t red - redirect\n"
					     "\t ts - timestamp request\n"
					     "\t info - info request\n"
					     "\t param - param problem\n");
			else
				mthis->usage("type not supported\n"
					     "Currently supported types are:\n"
					     "\t echo - echo request\n"
					     "\t du - dest unreachable\n"
					     "\t mask - mask request\n"
					     "\t red - redirect\n"
					     "\t ts - timestamp request\n"
					     "\t info - info request\n"
					     "\t param - param problem\n");	
			if(!ICMP_INFOTYPE(my_type) && my_type != ICMP_PARAMPROB 
			   && my_type != ICMP_UNREACH && my_type != ICMP_REDIRECT)
				mthis->usage("type not supported");
			break;
	        case 'c':
			mlen = strlen(optarg);
			my_code = atoi(optarg);
			break;
			
                default:
                        //we dont handle anything else
                        return PLUGIN_ERROR;
        }
        // we handled it
        return PLUGIN_OK;
}

/** @fn icmp_leadin(void *c_data, int i_data)
  * Plugin Start function
  * @param c_data Not used yet
  * @param i_data Not used yet
  * @return PLUGIN_OK
  * @sideeffect Plugin will be ready to sweep
  */
static int icmp_leadin(void *c_data, int i_data)
{
        // load socket stuff here
        // i_data will be the TTL to set it to
        uid_t uid;
        struct linger l;
        {
                assert(mthis != NULL);
                assert(mthis->usage != NULL);
                assert(mthis->getprivs != NULL);
                assert(mthis->dropprivs != NULL);
        }
        if(delay == -1)
                delay = 0;      //5 * 60; //default delay of 5 minutes
        // setuid root section of code
        {
                uid = mthis->getprivs();
                // root section of code
                //get me a RAW ICMP socket
                my_socket = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
                if(my_socket == -1)
                        mthis->usage("could not create socket");
                if(setsockopt(my_socket, SOL_IP, IP_TTL, &(ttl),
                              sizeof(ttl)) < 0)
                        mthis->usage("could not set TTL");
                l.l_onoff = 1;
                l.l_linger = delay;
                if(setsockopt(my_socket, SOL_SOCKET, SO_LINGER,
                              (struct linger *) &l,
                              sizeof(struct linger)) < 0)
                        mthis->usage("could not set linger time");
                mthis->dropprivs(uid);
        }
        //back to normalcy
        ready = true;
        seq_no = 1 + (int) (70.0 * rand() / (RAND_MAX + 1.0));

        return PLUGIN_OK;
}

/** @fn icmp_close(void)
  * Plugin destroy function
  * Call this before icmp_end
  * @return PLUGIN_OK
  * @sideeffect Plugin will be stopped and not ready
  */
static int icmp_close(void)
{
        close(my_socket);
        my_socket = -1;
        ready = false;
        ttl = 0;
        delay = -1;
        size = 0;
        payload_file = NULL;
        pthis.ready = false;
        return PLUGIN_OK;
}

/** @fn icmp_sweep(void *data)
  * Plugin Sweep Function
  * Runs a network sweep (this is for network plugins only).
  * Recompiled to make one pass only!!!
  * @param data Not used yet
  * @return PLUGIN_ERROR on error, PLUGIN_OK on ok
  */
static int icmp_sweep(void *data)
{
        unsigned int counter = 0;
        // just make sure we, and sonar, are ready for processing
        if(!ready || !pthis.ready)
        {
                return PLUGIN_ERROR;
        }

        //TODO: randomize host list

        // make my list circular, so this loop will never end
        hosts->next = first_host;
        hosts = first_host;

        while(hosts != NULL && counter != num_hosts)
        {
                if(hosts->ip == 0)
                {
                        //this is the "header"
                        hosts = hosts->next;
                        continue;
                }
                if(icmp_ping(hosts) == PLUGIN_ERROR)
                        return PLUGIN_ERROR;
                if(icmp_recv(hosts) == PLUGIN_NORESP)
                {
                        mthis->message("no response from: %s\n", hosts->host);
                }
                else
                {
                        if(cmd != NULL)
                        {
                                mthis->message("executing \"%s\"\n", cmd);
                                //we dont wanna run this as root... that
                                //would be possibly bad, so if they want
                                //root access, they need to sudo or su root
                                system(cmd);
                        }
                }
                hosts = hosts->next;
                counter++;
        }
        sleep(delay);
        return PLUGIN_OK;
}

/** @fn icmp_status(void)
  * Plugin Status hook function
  * Prints out status information right before the sweep begins
  */
static void icmp_status(void)
{
        if(pthis.ready)
        {
                mthis->output("ICMP sweep: %d host(s)::\n", num_hosts);
		mthis->output("\tType: %s Code: %d (non-verified)\n",
			      respstr(my_type), my_code);
                if(delay == 0)
                        mthis->output("\tdefault delay::\n");
                else
                        mthis->output("\tdelay %ds::\n", delay);
                if(payload_file != NULL)
                        mthis->output("\tpayload: %s::\n", payload_file);
                else
                        mthis->output("\tno payload::\n");
                if(size != 0)
                        mthis->output("\tsize %d::\n", size);
                else
                        mthis->output("\trandom packet size::\n");
                if(ttl != 255)
                        mthis->output("\tTTL %d::\n", ttl);
                else
                        mthis->output("\tdefault TTL::\n");
                if(cmd != NULL)
                        mthis->output("\tcommand: \"%s\"\n", cmd);
        }
}

/** @fn icmp_usage(void)
  * Plugin Usage hook function
  * Prints out plugin usage information. All options preceded by -p are passed
  * off to plugins, so make your usage reflect that.  However, when your arguments
  * get to your plugin's args function, the p will be stripped.
  */
static void icmp_usage(void)
{
        {
                assert(mthis->output != NULL);
        }
        mthis->output(" ICMP Scan plugin: requires root priveleges\n");
        mthis->output
                ("  Usage: [-pp file] [-ps size] [-pt ttl] [-pe cmd] [-pi type] [-pc code] hostname...\n");
        mthis->output("\t  -pd: delay for sec seconds after sweeps.\n");
        mthis->output
                ("\t  -pp: append contents of file to payload of ICMP packet\n");
        mthis->output("\t  -ps: set packet size\n");
        mthis->output("\t  -pt: set packet ttl\n");
        mthis->output
                ("\t  -pe: execute cmd if host is online (will NOT be run as root)\n");
	mthis->output("\t  -pi: set ICMP type (-pihelp for a list)\n");
	mthis->output("\t  -pc: set ICMP code\n");
        mthis->output("  append any number of hostnames...\n");
        mthis->output("   Prepend 'I' to all hostnames: Iwww.google.com\n");
}

//@}

/** @name Internal functions
  * Every plugin you make must have some internal helper functions
  * It might be a good idea to declare these static
 */
//@{

/** @fn icmp_ping(host_list_t *hosts)
  * Plugin Ping function
  * Sends out an ICMP ECHO_REQUEST packet
  * @param hosts pointer to a host_list_t structure containing current host to ping
  * @return PLUGIN_OK on ok, PLUGIN_ERROR otherwise
  * @todo Multithread this and the recv function
  */
static int icmp_ping(host_list_t * hosts)
{
        icmphdr_ *icmp_hdr;
        int total_size = sizeof(icmphdr_), dlen = 0;
        int bwrote;
        //struct timezone tz;
        //register struct timeval *tp;
        char *buf, *datapart;
        struct sa dest;

        {
                assert(hosts != NULL);
                assert(hosts->host != NULL);
                assert(hosts->ip != 0);
                assert(hosts->ip != (unsigned int) -1);
                assert(my_socket != 0);
                assert(my_socket != -1);
        }

        if(size != 0)
                total_size += size;
        else
                total_size += 64 + (int) (290.0 * rand() / (RAND_MAX + 64.0));
        dlen = total_size - sizeof(icmphdr_);
        buf = (char *) malloc(total_size);
        if(!buf)
                return PLUGIN_ERROR;
        memset(buf, 0, total_size);

        //this will overwrite some data;
        //tp = (struct timeval *) (buf + sizeof(icmphdr_));

        icmp_hdr = (icmphdr_ *) buf;
        icmp_hdr->i_type = my_type; //ICMP_ECHO;
        icmp_hdr->i_code = my_code; //0;
        icmp_hdr->i_id = (unsigned short) getpid();
        icmp_hdr->i_cksum = 0;
        icmp_hdr->i_seq = seq_no++;

        dest.family = PF_INET;
        dest.dport = 0;
        dest.addr = hosts->ip;
        memset(dest.zero, 0, 8 * sizeof(unsigned char));

        datapart = buf + sizeof(icmphdr_);
        if(!payload_file)
        {
                memset(datapart, (int) (256.0 * rand() / (RAND_MAX + 0.0)),
                       dlen);
        }
        else
        {
                unsigned int len = 0, cur = 0;
                FILE *payload;
                payload = fopen(payload_file, "r");
                if(!payload)
                        return PLUGIN_ERROR;
                fseek(payload, 0, SEEK_END);
                len = (unsigned int) ((ftell(payload) > dlen) ? dlen :
                                      ftell(payload));
                fseek(payload, 0, SEEK_SET);
                while(!feof(payload) && cur < len)
                {
                        int c = fgetc(payload);
                        datapart[cur++] = c;
                }
        }

        gettimeofday(&tp, &(tz));

        ((icmphdr_ *) buf)->i_cksum =
                checksum((unsigned short *) buf, total_size);
        bwrote = sendto(my_socket, buf, total_size, 0,
                        (struct sockaddr *) &(dest), sizeof(dest));
        if(bwrote == -1)
        {
                error("sendto failed: bailing... \n", PLUGIN_ERROR);
                perror("sendto");
                return PLUGIN_ERROR;
        }
        if(bwrote < (int) total_size)
        {
                mthis->message("wrote %d bytes\n", bwrote);
        }
        free(buf);
        return PLUGIN_OK;
}

/** @fn icmp_recv(host_list_t *hosts)
  * Plugin Receive function
  * Looks for an ICMP ECHO_REPLY packet
  * @param hosts pointer to a host_list_t structure containing current host to ping
  * @return PLUGIN_RESP if packet received, PLUGIN_NORESP otherwise 
  * @return PLUGIN_ERROR on error, PLUGIN_OK on other data recv'd
  * @todo Multithread this and the recv function
  */
static int icmp_recv(host_list_t * hosts)
{
        fd_set rfds;
        iphdr_ *iphdr;
        icmphdr_ *icmphdr;
        unsigned short iphdrlen;
        unsigned long triptime;
        struct timeval tv;//, *tp;
        //struct timezone tz;
        int bread, ret;
        char buf[MAX_PACKET];
        int fromlen;
        struct sockaddr_in from;

        if(delay > 0)
                tv.tv_sec = delay;
        else
                tv.tv_sec = 5;
        tv.tv_usec = 0;

      loop:
        FD_ZERO(&rfds);
        FD_SET(my_socket, &rfds);
        ret = select(my_socket + 1, &rfds, NULL, NULL, &tv);
        if(ret == -1)
        {
                error("could not select socket\n", PLUGIN_ERROR);
                return PLUGIN_ERROR;
        }
        if(ret == 0)
        {
                return PLUGIN_NORESP;
        }

        bread = recvfrom(my_socket, buf, MAX_PACKET, 0,
                         (struct sockaddr *) &(from),
                         (socklen_t *) & (fromlen));
        if(bread == -1)
        {
                mthis->message("recvfrom error\n");
		perror("recvfrom");
                return PLUGIN_ERROR;
        }
        iphdr = (iphdr_ *) buf;
        iphdrlen = iphdr->h_len * 4;    // number of 32-bit words *4 = bytes

        if(bread < iphdrlen + ICMP_MIN)
                mthis->message("Too few bytes from %s\n",
                               inet_ntoa(from.sin_addr));

        icmphdr = (icmphdr_ *) (buf + iphdrlen);

        if(icmphdr->i_type != resptype(my_type))
        {
                mthis->message("non-%s recv'd - type: %s code: ",
                               respstr(resptype(my_type)), respstr(icmphdr->i_type));
		if(icmphdr->i_type == ICMP_UNREACH)
			mthis->output("%s(%d) ", unreach_codes(icmphdr->i_code), icmphdr->i_code);
		else
			mthis->output("%d ", icmphdr->i_code);
		mthis->output("id: %d\n", icmphdr->i_id);
                return PLUGIN_OK;
        }
        if(icmphdr->i_id != (unsigned short) getpid())
        {
                //keep looking
                goto loop;
        }
        gettimeofday(&tv, &(tz));

        tvsub(&tv, &tp);
        triptime = tv.tv_sec * 1000 + (tv.tv_usec / 1000);
        mthis->message
                ("%s[%s] online: icmp_seq = %d. range: %lu ms ttl: %d size: %d\n",
                 hosts->host, inet_ntoa(from.sin_addr), icmphdr->i_seq,
                 triptime, iphdr->ttl, ntohs(iphdr->total_len));
        return PLUGIN_RESP;
}

/**
  * @fn checksum(unsigned short *buffer, int size)
  * Compute an ICMP checksum
  * @param buffer Pointer to data
  * @param size Size of data
  * @return Checksum of data
  */
static unsigned short checksum(unsigned short *buffer, int size)
{

        unsigned long cksum = 0;

        while(size > 1)
        {
                cksum += *buffer++;
                size -= sizeof(unsigned short);
        }

        if(size)
        {
                cksum += *(unsigned char *) buffer;
        }

        cksum = (cksum >> 16) + (cksum & 0xffff);
        cksum += (cksum >> 16);
        return (unsigned short) (~cksum);
}

/**
  * @fn tvsub(register struct timeval *out, register struct timeval *in)
  * Subtract one struct timeval from another
  * @param out Pointer to destination (will be filled with out - in)
  * @param in Pointer to src
*/
static void tvsub(register struct timeval *out, register struct timeval *in)
{
        if((out->tv_usec -= in->tv_usec) < 0)
        {
                out->tv_sec--;
                out->tv_usec += 1000000;
        }
        out->tv_sec -= in->tv_sec;
}

static void icmp_escape_shell(char func[])
{
        int len = strlen(func);
        int i = 0;
        for(i = 0; i < len; i++)
        {
                if(func[i] == '>' || func[i] == '<' ||
                   func[i] == ';' || func[i] == '(' ||
                   func[i] == '`' || func[i] == '!' ||
                   func[i] == '|' || func[i] == '&' ||
                   func[i] == '*' || func[i] == '?' ||
                   func[i] == '$' || func[i] == ':' ||
                   func[i] == '=' || func[i] == '#' ||
                   func[i] == '!' || func[i] == '%' ||
                   func[i] == ')' || func[i] == '[' ||
                   func[i] == ']' || func[i] == '{' || func[i] == '}' ||
                   /* somone trying to pass shellcode w/nop's? */
                   func[i] == (char) 0x90)
                        //we want to get rid of these kinds of characters
                        //so, move the string forward
                {
                        memmove(&func[i], &func[i + 1], len - (i + 1));
                        func[len-- - 1] = 0x00;
                }
        }
}

static int icmp_last_error(void)
{
        return my_errno;
}

static void inline error(char *x, int y)
{
	mthis->message("%s:%d ERROR %s\n", __FILE__, __LINE__, x);
	*(pthis.errno) = y;
}

static int resptype(int type)
{
	switch(type) 
	{
	case ICMP_ROUTERADVERT:
		return ICMP_ROUTERSOLICIT;
	case ICMP_ROUTERSOLICIT:
		return ICMP_ROUTERADVERT;	
	case ICMP_TSTAMP:
		return ICMP_TSTAMPREPLY;
	case ICMP_IREQ:
		return ICMP_IREQREPLY;
	case ICMP_MASKREQ:
		return ICMP_MASKREPLY;
	case ICMP_ECHO:
		return ICMP_ECHOREPLY;
	}
	return -1;
}

static const char * respstr(int type)
{
	switch(type) 
	{		
	case ICMP_UNREACH:
		return "dest unreachable";
	case ICMP_SOURCEQUENCH:
		return "source quench";
	case ICMP_ROUTERADVERT:
		return "router advert";
	case ICMP_ROUTERSOLICIT:
		return "router solicit";
	case ICMP_TIMXCEED:
		return "time exceed";
	case ICMP_PARAMPROB:
		return "param problem";
	case ICMP_TSTAMP:
		return "timestamp request";
	case ICMP_TSTAMPREPLY:
		return "timestamp reply";
	case ICMP_IREQ:
		return "info request";
	case ICMP_IREQREPLY:
		return "info reply";
	case ICMP_MASKREQ:
		return "mask request";
	case ICMP_MASKREPLY:
		return "mask reply";
	case ICMP_ECHO:
		return "echo request";
	case ICMP_ECHOREPLY:
		return "echo reply";
	}
	return "unknown";
}

static const char * unreach_codes(int code)
{
	switch(code)
	{
				// unreach codes:
	case ICMP_UNREACH_NET:
		return "net unreachable";
	case ICMP_UNREACH_HOST:
		return "host unreachable";
	case ICMP_UNREACH_PROTOCOL:
		return "protocol not supported";
	case ICMP_UNREACH_PORT:
		return "port unreachable";
	case ICMP_UNREACH_NEEDFRAG:
		return "need fragmentation";
	case ICMP_UNREACH_SRCFAIL:
		return "source-route failed";
	case ICMP_UNREACH_NET_UNKNOWN:
		return "network unknown";
	case ICMP_UNREACH_HOST_UNKNOWN:
		return "host unknown";
	case ICMP_UNREACH_ISOLATED:
		return "host isolated";
	case ICMP_UNREACH_NET_PROHIB:
		return "net denied";
	case ICMP_UNREACH_HOST_PROHIB:
		return "host denied";
	case ICMP_UNREACH_TOSNET:
		return "bad TOS for net";
	case ICMP_UNREACH_TOSHOST:
		return "bad TOS for host";
	case ICMP_UNREACH_FILTER_PROHIB:
		return "filter prohibited";
	case ICMP_UNREACH_HOST_PRECEDENCE:
		return "host precedence violation";
	case ICMP_UNREACH_PRECEDENCE_CUTOFF:
		return "precedence cutoff";
	}
	return "unknown";
}

//@}
