/***************************************************************************
   network_icmp.h  -  ICMP network plugin
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: network_icmp.h,v 1.15 2002/08/22 21:54:37 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file network_icmp.h
  * File contains ICMP plugin functions
  * This file is a good example of the plugin API.
  * Use it as an example for your own plugin development.
  */
#ifndef _ICMP_PLUG_H_
#define _ICMP_PLUG_H_

// include the plugin API
#include "plugin.h"

/* Get all the types first */
#include <sys/types.h>
#include <netinet/in_systm.h>

#ifdef MACOSX
#ifndef socklen_t
typedef int socklen_t;
#endif
#endif

#define LINUX 1

#ifndef SOL_IP
#define SOL_IP 0
#endif

#ifndef INADDR_NONE
#define	INADDR_NONE		((in_addr_t) 0xffffffff)
#endif

// include my types
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>

/// Plugin's plugin_t structure, built by plugin_init
plugin_t pthis;
/// Sonar's sonar_t structure, built by sonar, passed to this plugin
sonar_t *mthis;

/** @name ICMP definitions
  * These are standard ICMP definitions
 */
//@{
#define ICMP_ECHO 	8
#define ICMP_ECHOREPLY 	0
#define ICMP_MIN 	8
//@}

/// Maximum receivable packet size
#define MAX_PACKET 1024

/// This plugin's debug script
#define DEBUG2(x) 	mthis->output("%s:%d: %s\n", __FILE__, __LINE__, x);


/// my last error
static int my_errno;
/// my socket
static int my_socket,
/// my ttl
    ttl,
/// my delay
    delay,
/// my size
    size,
/// my sequence number
    seq_no,
/// my ICMP type
    my_type,
/// my ICMP code
    my_code;

/// ready for use
static bool ready;
/// payload file name
static char *payload_file;
/// number of targets
static unsigned int num_hosts;
/// Timestamp Time zone
struct timezone tz;
/// Timestamp timeval
struct timeval tp;

/**
  * @struct _iphdr
  * IP Packet header
  */
typedef struct _iphdr
{
#ifndef WORDS_BIGENDIAN
    /** Header length */
	unsigned char h_len:4,
    /** Version */
		version:4;
#else
    /** Version */
	unsigned char version:4,
    /** Header length */
		h_len:4;
#endif
    /** Type of Service */
	unsigned char tos;
    /** Total Packet Length */
	unsigned short total_len,
    /** ID Number */
		ident,
    /** Offset to Data */
		offset;
    /** Time to Live */
	unsigned char ttl,
    /** Protocol Number */
		proto;
    /** Checksum */
	unsigned short checksum;
    /** Source IP Address */
	unsigned int sourceIP,
    /** Dest IP Address */
		destIP;
}iphdr_;


/**
  * @struct _icmphdr
  * ICMP Packet header
  */
typedef struct _icmphdr
{
  /** ICMP Type */
	unsigned char i_type;
  /** ICMP Type SubCode */
	unsigned char i_code;
  /** ICMP Checksum */
	unsigned short i_cksum;
  /** ICMP ID */
	unsigned short i_id;
  /** ICMP Sequence */
	unsigned short i_seq;
}icmphdr_;

/**
  * @struct __host_list_t
  * Circular host queue, to avoid a memory leak
  */

typedef struct __host_list_t
{
	/// next in list
	struct __host_list_t *next;
	/// this one's hostname
	char *host;
	/// this one's IP
	unsigned int ip;
}host_list_t;

/// current host pointer
host_list_t *hosts;
/// header host pointer
host_list_t *first_host;

void *plugin_init(void *in_data);
static int icmp_end(void);
static int icmp_add(char *hostname);
static int icmp_args(char opt, char *optarg);
static int icmp_leadin(void *c_data, int i_data);
static int icmp_close(void);
static int icmp_sweep(void *data);
static void icmp_usage(void);
static void icmp_status(void);
static int icmp_ping(host_list_t * hosts);
static int icmp_recv(host_list_t * hosts);
static unsigned short checksum(unsigned short *buffer, int size);
static void tvsub(register struct timeval *out,
		  register struct timeval *in);
static void icmp_escape_shell(char func[]);
static int icmp_last_error(void);
static void inline error(char *x, int y);
static int resptype(int type);
static const char * respstr(int type);
static const char * unreach_codes(int code);

#endif                          //_ICMP_PLUG_H_
