/***************************************************************************
   passive.c  -  passive scanning plugin
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: passive.c,v 1.6 2002/08/22 21:55:19 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file passive.c
  * Contains a passive scanning plugin that listens for incoming traffic and
  * displays the hostname and IP address of incoming packets
  */

#include <assert.h>
#include "passive.h"

static const char rcsid[] =
        "\100$ autosec/sonar: $Id: passive.c,v 1.6 2002/08/22 21:55:19 red0x Exp $";

void *plugin_init(void *in_data)
{
        {
                assert(in_data != NULL);
        }
        //useless reference, to avoid a warning
        if(rcsid)
        {
        }
        // this is assumed
        mthis = (sonar_t *) in_data;

        pthis.type = NETWORK_PLUGIN;
        pthis.errno = &my_errno;
        pthis.ready = false;
        pthis.plugin_init = &plugin_init;
        pthis.plugin_destroy = &plugin_destroy;
        pthis.plugin_args = &plugin_args;
        pthis.plugin_open = &plugin_open;
        pthis.plugin_close = &plugin_close;
        pthis.plugin_next = &plugin_next;
        // this is not needed for I/O plugins
        pthis.plugin_sweep = &plugin_sweep;
        // red0x: do we even need this??
        pthis.plugin_go = NULL;
        pthis.plugin_usage = &plugin_usage;
        pthis.plugin_status = &plugin_status;
        pthis.plugin_last_error = &plugin_last_error;
        my_errno = PLUGIN_LOADED;
        my_delay = 5;

        return &pthis;
}


static int plugin_destroy(void)
{

        mthis = NULL;

        pthis.type = 0;
        pthis.errno = NULL;
        pthis.ready = false;
        pthis.plugin_init = NULL;
        pthis.plugin_destroy = NULL;
        pthis.plugin_open = NULL;
        pthis.plugin_close = NULL;
        pthis.plugin_next = NULL;
        pthis.plugin_sweep = NULL;
        pthis.plugin_args = NULL;
        pthis.plugin_sweep = NULL;
        // red0x: do we even need this??
        pthis.plugin_go = NULL;
        pthis.plugin_usage = NULL;
        pthis.plugin_status = NULL;
        pthis.plugin_last_error = NULL;
        my_errno = PLUGIN_UNLOADED;
        return my_errno;
}

static int plugin_args(char opt, char *optarg)
{
        assert(mthis != NULL);
        assert(mthis->output != NULL);
        assert(mthis->memput != NULL);
        switch (opt)
        {
                case 'd':
                        my_delay = atoi(optarg);
                        break;
                default:
                        //we dont handle this
                        return PLUGIN_ERROR;
        }
        return PLUGIN_OK;
}

static int plugin_open(void *c_data, int i_data)
{
        uid_t uid;
        struct linger l;
        assert(mthis != NULL);
        assert(mthis->usage != NULL);
        assert(mthis->getprivs != NULL);
        assert(mthis->dropprivs != NULL);
        //setuid section of code
        {
                uid = mthis->getprivs();
                // root section of code
                //get me  RAW TCP and ICMP sockets
                my_socket = socket(PF_INET, SOCK_RAW, IPPROTO_TCP);
                if(my_socket == -1)
                {
                        perror("socket");
                        mthis->usage("could not create raw TCP socket");
                }

                l.l_onoff = 1;
                l.l_linger = my_delay;
                if(setsockopt(my_socket, SOL_SOCKET, SO_LINGER,
                              (struct linger *) &l,
                              sizeof(struct linger)) < 0)
                        mthis->usage
                                ("could not set linger time on TCP socket");

                myudp_socket = socket(PF_INET, SOCK_RAW, IPPROTO_UDP);
                if(my_socket == -1)
                {
                        perror("socket");
                        mthis->usage("could not create raw UDP socket");
                }

                l.l_onoff = 1;
                l.l_linger = my_delay;
                if(setsockopt(myudp_socket, SOL_SOCKET, SO_LINGER,
                              (struct linger *) &l,
                              sizeof(struct linger)) < 0)
                        mthis->usage
                                ("could not set linger time on UDP socket");

                myicmp_socket = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
                if(my_socket == -1)
                {
                        perror("socket");
                        mthis->usage("could not create raw ICMP socket");
                }

                l.l_onoff = 1;
                l.l_linger = my_delay;
                if(setsockopt(myicmp_socket, SOL_SOCKET, SO_LINGER,
                              (struct linger *) &l,
                              sizeof(struct linger)) < 0)
                        mthis->usage
                                ("could not set linger time on ICMP socket");
                mthis->dropprivs(uid);
        }
        //back to normalcy
        pthis.ready = true;

        return PLUGIN_OK;
}

static int plugin_close(void)
{
        close(my_socket);
        pthis.ready = false;
        my_delay = -1;

        return PLUGIN_OK;
}

static int plugin_next(char *hostname)
{
        char t;
        assert(mthis != NULL);
        assert(mthis->memput != NULL);
        assert(mthis->resolver != NULL);

        if(hostname == NULL)
                return PLUGIN_ERROR;

        t = (char) *hostname;
        if(t != 'L')
                return PLUGIN_ERROR;

        return plugin_open(NULL, 0);

        //return PLUGIN_OK;
}

static void plugin_usage(void)
{
        assert(mthis != NULL);
        assert(mthis->output != NULL);
        mthis->output("\n Passive Scanning plugin: requires root priveleges\n");
        mthis->output("\t \'sonar L\' will fire up the passive scanner\n");

}

static void plugin_status(void)
{
        assert(mthis != NULL);
        assert(mthis->output != NULL);
        if(pthis.ready)
        {
                mthis->output("Passive scanner:: ready\n");
        }
}

static int plugin_last_error(void)
{
        return my_errno;
}

static int plugin_sweep(void *data)
{
        char packet[BUFSIZE], *hostname;
        int bread, fromlen, ret, ss;
        unsigned short iphdrlen;
        struct tcphdr *tcp;
        struct udphdr *udp;
        struct icmp *icmp;
        struct ip *ip;
        struct sockaddr_in from;
        struct timeval tv;
        struct hostent *he;
        //      struct sa tp;
        fd_set rfds;

        assert(mthis != NULL);
        assert(mthis->output != NULL);
        assert(my_socket != -1 && my_socket != 0);

        tv.tv_sec = my_delay;
        tv.tv_usec = 0;

      loop:
        memset(packet, 0, BUFSIZE);
        FD_ZERO(&rfds);
        FD_SET(my_socket, &rfds);
        FD_SET(myudp_socket, &rfds);
        FD_SET(myicmp_socket, &rfds);
        if(myudp_socket > my_socket)
                ss = myudp_socket;
        else
                ss = my_socket;
        if(myicmp_socket > ss)
                ss = myicmp_socket;

        ret = select(ss + 1, &rfds, NULL, NULL, &tv);
        if(ret == -1)
        {
                mthis->message("could not select sockets\n");
                return PLUGIN_ERROR;
        }
        if(ret == 0)
        {
                goto loop;
        }
        if(FD_ISSET(my_socket, &rfds))
        {
                bread = recvfrom(my_socket, packet, BUFSIZE, 0,
                                 (struct sockaddr *) &(from),
                                 (socklen_t *) & (fromlen));
                if(bread == -1)
                {
                        perror("recvfrom");
                        mthis->message("recvfrom error\n");
                        return PLUGIN_ERROR;
                }
                ip = (struct ip *) packet;
                iphdrlen = ip->ip_hl * 4;
                if(bread < iphdrlen)
                {
                        mthis->message("undersized packet from %s\n",
                                       inet_ntoa(from.sin_addr));
                        goto loop;
                }

                //reverse resolve the IP to name
                he = gethostbyaddr((const char*) &(from.sin_addr), 4, AF_INET);
                if(he)
                        mthis->memput(&hostname, he->h_name,
                                      strlen(he->h_name));
                else
                        mthis->memput(&hostname, inet_ntoa(from.sin_addr),
                                      strlen(inet_ntoa(from.sin_addr)));

                print_ip(ip, from, hostname);

                tcp = (struct tcphdr *) (packet + iphdrlen);
                print_tcp(tcp, from);
                free(hostname);
                goto loop;
        }
        if(FD_ISSET(myudp_socket, &rfds))
        {
                bread = recvfrom(myudp_socket, packet, BUFSIZE, 0,
                                 (struct sockaddr *) &(from),
                                 (socklen_t *) & (fromlen));
                if(bread == -1)
                {
                        perror("recvfrom");
                        mthis->message("recvfrom error\n");
                        return PLUGIN_ERROR;
                }
                ip = (struct ip *) packet;
                iphdrlen = ip->ip_hl * 4;
                if(bread < iphdrlen)
                {
                        mthis->message("undersized packet from %s\n",
                                       inet_ntoa(from.sin_addr));
                        goto loop;
                }

                //reverse resolve the IP to name
                he = gethostbyaddr((const char*) &(from.sin_addr), 4, AF_INET);
                if(he)
                        mthis->memput(&hostname, he->h_name,
                                      strlen(he->h_name));
                else
                        mthis->memput(&hostname, inet_ntoa(from.sin_addr),
                                      strlen(inet_ntoa(from.sin_addr)));

                print_ip(ip, from, hostname);

                udp = (struct udphdr *) (packet + iphdrlen);
                print_udp(udp, from);
                free(hostname);
                goto loop;
        }
        if(FD_ISSET(myicmp_socket, &rfds))
        {
                bread = recvfrom(myicmp_socket, packet, BUFSIZE, 0,
                                 (struct sockaddr *) &(from),
                                 (socklen_t *) & (fromlen));
                if(bread == -1)
                {
                        perror("recvfrom");
                        mthis->message("recvfrom error\n");
                        return PLUGIN_ERROR;
                }
                ip = (struct ip *) packet;
                iphdrlen = ip->ip_hl * 4;
                if(bread < iphdrlen)
                {
                        mthis->message("undersized packet from %s\n",
                                       inet_ntoa(from.sin_addr));
                        goto loop;
                }

                //reverse resolve the IP to name
                he = gethostbyaddr((const char *) &(from.sin_addr), 4, AF_INET);
                if(he)
                        mthis->memput(&hostname, he->h_name,
                                      strlen(he->h_name));
                else
                        mthis->memput(&hostname, inet_ntoa(from.sin_addr),
                                      strlen(inet_ntoa(from.sin_addr)));

                print_ip(ip, from, hostname);

                icmp = (struct icmp *) (packet + iphdrlen);
                print_icmp(icmp, from);
                free(hostname);
                goto loop;
        }

        return PLUGIN_OK;
}

static void print_ip(struct ip *ip, struct sockaddr_in sa, char *hostname)
{
        assert(mthis != NULL);
        assert(mthis->message != NULL);
        assert(mthis->output != NULL);
        mthis->message("%s[%s] online::", hostname, inet_ntoa(sa.sin_addr));
        //print the packet stuff
}

static void print_udp(struct udphdr *udp, struct sockaddr_in sa)
{
        assert(mthis != NULL);
        assert(mthis->message != NULL);
        assert(mthis->output != NULL);
        mthis->output("  UDP/IP packet:: (%d) -> (%d)\n", ntohs(udp->uh_sport),
                      ntohs(udp->uh_dport));
}

static void print_tcp(struct tcphdr *tcp, struct sockaddr_in sa)
{
        assert(mthis != NULL);
        assert(mthis->message != NULL);
        assert(mthis->output != NULL);
        mthis->output("  TCP/IP packet:: (%d) -> (%d)\n", ntohs(tcp->th_sport),
                      ntohs(tcp->th_dport));
}

static void print_icmp(struct icmp *icmp, struct sockaddr_in sa)
{
        assert(mthis != NULL);
        assert(mthis->message != NULL);
        assert(mthis->output != NULL);
        mthis->output("  ICMP/IP packet::\n");
}
