/***************************************************************************
   rfc793.cpp  -  TCP port 0 plugin
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: rfc793.c,v 1.7 2002/08/22 21:56:22 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file rfc793.c
  * Contains a rfc 793 behavior check (TCP port 0 ping)
  */

#include <assert.h>
#include "rfc793.h"

static const char rcsid[] =
        "\100$ autosec/sonar: $Id: rfc793.c,v 1.7 2002/08/22 21:56:22 red0x Exp $";

/** @fn plugin_init(void *in_data)
  * Plugin Init function, part of the Plugin API
  * @param in_data Pointer to a sonar_t structure that will be 
  *  passed from sonar when your plugin is loaded.
  * @return A pointer to a filled plugin_t structure.
  */

static int my_socket, err_socket, delay, ttl, port;
static unsigned int num_hosts;

void *plugin_init(void *in_data)
{
        {
                assert(in_data != NULL);
        }
        //useless reference, to avoid a warning
        if(rcsid)
        {
        }
        // this is assumed
        mthis = (sonar_t *) in_data;

        pthis.type = NETWORK_PLUGIN;
        pthis.errno = &my_errno;
        pthis.ready = false;
        pthis.plugin_init = &plugin_init;
        pthis.plugin_destroy = &plugin_destroy;
        pthis.plugin_args = &plugin_args;
        pthis.plugin_open = &plugin_open;
        pthis.plugin_close = &plugin_close;
        pthis.plugin_next = &plugin_next;
        pthis.plugin_sweep = &rfc793_sweep;
        // red0x: do we even need this??
        pthis.plugin_go = NULL;
        pthis.plugin_usage = &plugin_usage;
        pthis.plugin_status = &plugin_status;
        pthis.plugin_last_error = &plugin_last_error;

        hosts = (host_list_t *) malloc(sizeof(host_list_t));
        first_host = hosts;
        if(!hosts)
        {
                my_errno = PLUGIN_UNLOADED;
                return NULL;
        }
        hosts->next = NULL;
        hosts->host = NULL;
        my_errno = PLUGIN_LOADED;

        my_socket = port = -1;
        delay = 0;
        ttl = 255;

        num_hosts = 0;

        return &pthis;
}


static int plugin_destroy(void)
{
        mthis = NULL;
        if(my_socket > 0)
        {
                close(my_socket);
                my_socket = -1;
        }

        pthis.type = 0;
        pthis.errno = NULL;
        pthis.ready = false;
        pthis.plugin_init = NULL;
        pthis.plugin_destroy = NULL;
        pthis.plugin_open = NULL;
        pthis.plugin_close = NULL;
        pthis.plugin_next = NULL;
        pthis.plugin_sweep = NULL;
        pthis.plugin_args = NULL;
        pthis.plugin_sweep = NULL;
        // red0x: do we even need this??
        pthis.plugin_go = NULL;
        pthis.plugin_usage = NULL;
        pthis.plugin_status = NULL;
        pthis.plugin_last_error = NULL;
        my_errno = PLUGIN_UNLOADED;
        my_socket = delay = ttl = port = -1;
        if(hosts)
                free(hosts);
        hosts = NULL;

        return my_errno;
}

static int plugin_args(char opt, char *optarg)
{
        assert(mthis != NULL);
        assert(mthis->output != NULL);
        assert(mthis->memput != NULL);
        switch (opt)
        {
                case 'd':
                        delay = atoi(optarg);
                        if(delay < 0)
                                delay = 0;
                        break;
                case 't':
                        ttl = atoi(optarg);
                        if(ttl > 255)
                                ttl = 255;
                        if(ttl < 1)
                                ttl = 255;
                        break;
                case 'p':
                        port = atoi(optarg);
                        if(port < 0)
                                port = -1;
                        if(port > 65535)
                                port = -1;
                        break;
                default:
                        //we dont handle this
                        return PLUGIN_ERROR;
        }
        return PLUGIN_OK;
}

static int plugin_open(void *c_data, int i_data)
{
        // load socket stuff here
        // i_data will be the TTL to set it to
        uid_t uid;
        struct linger l;
        {
                assert(mthis != NULL);
                assert(mthis->usage != NULL);
                assert(mthis->getprivs != NULL);
                assert(mthis->dropprivs != NULL);
        }

        // setuid root section of code
        {
                uid = mthis->getprivs();
                // root section of code
                //get me  RAW TCP and ICMP sockets
                my_socket = socket(PF_INET, SOCK_RAW, IPPROTO_TCP);
                if(my_socket == -1)
                        mthis->usage("could not create raw TCP socket");
                err_socket = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
                if(err_socket == -1)
                        mthis->usage("could not create raw ICMP socket");

                if(setsockopt(my_socket, SOL_IP, IP_TTL, &(ttl),
                              sizeof(ttl)) < 0)
                        mthis->usage("could not set TTL");
                l.l_onoff = 1;
                l.l_linger = delay;
                if(setsockopt(my_socket, SOL_SOCKET, SO_LINGER,
                              (struct linger *) &l,
                              sizeof(struct linger)) < 0)
                        mthis->usage
                                ("could not set linger time on TCP socket");
                if(setsockopt
                   (err_socket, SOL_SOCKET, SO_LINGER, (struct linger *) &l,
                    sizeof(struct linger)) < 0)
                        mthis->usage
                                ("could not set linger time on ICMP socket");
                mthis->dropprivs(uid);
        }
        //back to normalcy
        pthis.ready = true;

        return PLUGIN_OK;
}

static int plugin_close(void)
{
        close(my_socket);
        pthis.ready = false;
        ttl = 0;
        delay = -1;

        return PLUGIN_OK;
}

static int plugin_next(char *host_name)
{
        char t;
        char *hostname = host_name + sizeof(char);      // move past the 'A'
        char *range, *tmp;
        {
                assert(mthis != NULL);
                assert(mthis->memput != NULL);
                assert(hosts != NULL);
                assert(mthis->resolver != NULL);
        }
        //we want all hosts incoming to be prefixed with A
        t = (char) *host_name;
        if(t != 'A')
                return PLUGIN_ERROR;

        if((range = strchr((const char *) hostname, '[')) != NULL)
        {
                if((tmp = strchr((const char *) hostname, ']')) != NULL)
                {
                        //we have a valid range of the form "[1-255]" somewhere
                        // what we're gunna do is solve the first range, and let
                        // the recursion unroll the other ranges
                        //we need the character after the first [
                        char *host, *head, *zero, *tail, *h2, t;
                        int start, end, j, len;

                        // leave the 'A' prefix for now
                        hostname = host_name;

                        // save the precursor to the range                     
                        t = range[0];
                        range[0] = 0x00;
                        head = (char *) malloc(strlen(hostname));
                        strncpy(head, hostname, strlen(hostname));
                        range[0] = t;

                        // get the bound of the range
                        zero = strchr((const char *) range, '-');
                        if(zero == NULL)
                                mthis->usage("invalid range");
                        *zero = 0x00;
                        start = atoi((const char *) (range + sizeof(char)));
                        end = atoi((const char *) (zero + sizeof(char)));

                        // save everything after the range
                        len = strlen(tmp);
                        tail = (char *) malloc(len);
                        memset(tail, 0, len);
                        strncpy(tail, (const char *) (tmp + sizeof(char)),
                                len - 1);

                        // now we replace the '[%d-%d]' with a %d
                        host = (char *) malloc(strlen(head) + 3 +
                                               strlen(tail) + 5);
                        strncpy(host, head, strlen(head));
                        strncat(host, "%d", 2 * sizeof(char));
                        strncat(host, tail, strlen(tail));

                        for(j = start; j <= end; j++)
                        {
                                h2 = (char *) malloc(strlen(host) + 5);
                                memset(h2, 0, strlen(host) + 5);
                                snprintf(h2, strlen(host) + 5, host, j);

                                if(plugin_next(h2) != PLUGIN_OK)
                                        mthis->usage("invalid range");

                        }
                        return PLUGIN_OK;
                }
                else
                {
                        mthis->usage("invalid range");
                        //invalid range
                }
        }
        if((range = strchr((const char *) hostname, '*')) != NULL)
        {
                //we have a * range like 192.168.1.*
                int start = 1, end = 255, j, tlen, len;
                char *head, *tail, t, *host, *h2;

                // leave the 'A' intact
                hostname = host_name;

                // save the head
                t = range[0];
                range[0] = 0x00;
                head = (char *) malloc(strlen(hostname));
                strncpy(head, hostname, strlen(hostname));
                range[0] = t;

                //save the tail
                range++;
                tlen = strlen(range);
                tail = (char *) malloc(tlen + 1);
                memset(tail, 0, tlen + 1);
                strncpy(tail, range, tlen);

                //build the format string
                len = strlen(head) + strlen(tail) + (sizeof(char) * 2) + 5;
                host = (char *) malloc(len);
                memset(host, 0, len);
                strncpy(host, head, strlen(head));
                strncat(host, "%d", sizeof(char) * 2);
                strncat(host, tail, strlen(tail));
                for(j = start; j <= end; j++)
                {
                        h2 = (char *) malloc(strlen(host) + 5);
                        memset(h2, 0, strlen(host) + 5);
                        snprintf(h2, strlen(host) + 5, host, j);

                        if(plugin_next(h2) != PLUGIN_OK)
                                mthis->usage("invalid range");
                }

                return PLUGIN_OK;
        }
        if((range = strchr((const char *) hostname, '/')) != NULL)
        {
                int a, b, c, d, mask, len;
                char *host, *h2;
                range[0] = 0x00;
                range++;
                mask = atoi(range);

                if(mthis->ipverify(hostname) == 0)
                {
                        //we have a hostname
                        unsigned int ip = mthis->resolver(hostname);
                        struct in_addr tmp;
                        if(ip == INADDR_NONE)
                                mthis->usage("unable to resolve hostname");
                        tmp.s_addr = ip;
                        hostname = inet_ntoa(tmp);
                }
                switch (mask)
                {
                        case 0:
                                mthis->memput(&host, "A*.*.*.*\0",
                                              strlen("A*.*.*.*\0"));
                                if(plugin_next(host) != PLUGIN_OK)
                                        mthis->usage
                                                ("failure scanning whole internet");
                                return PLUGIN_OK;
                                break;
                        case 8:
                                mthis->memput(&host, "A%d.%d.%d.%d\0",
                                              strlen("A%d.%d.%d.%d\0"));
                                len = strlen(hostname) + 5;
                                h2 = (char *) malloc(len);
                                memset(h2, 0, len);
                                sscanf(hostname, "%d", &a);
                                for(b = 1; b <= 255; b++)
                                {
                                        for(c = 1; c <= 255; c++)
                                        {
                                                for(d = 1; d <= 255; d++)
                                                {
                                                        snprintf(h2, len,
                                                                 host, a, b,
                                                                 c, d);
                                                        if(plugin_next(h2) !=
                                                           PLUGIN_OK)
                                                                mthis->usage
                                                                        ("failure scanning class A");
                                                }
                                        }
                                }
                                return PLUGIN_OK;
                                break;
                        case 16:
                                mthis->memput(&host, "A%d.%d.%d.%d\0",
                                              strlen("A%d.%d.%d.%d\0"));
                                len = strlen(hostname) + 5;
                                h2 = (char *) malloc(len);
                                memset(h2, 0, len);
                                sscanf(hostname, "%d.%d", &a, &b);
                                for(c = 1; c <= 255; c++)
                                {
                                        for(d = 1; d <= 255; d++)
                                        {
                                                snprintf(h2, len, host, a, b,
                                                         c, d);
                                                if(plugin_next(h2) !=
                                                   PLUGIN_OK)
                                                        mthis->usage
                                                                ("failure scanning class B");
                                        }
                                }
                                return PLUGIN_OK;
                                break;
                        case 24:
                                mthis->memput(&host, "A%d.%d.%d.%d\0",
                                              strlen("A%d.%d.%d.%d\0"));
                                len = strlen(hostname) + 5;
                                h2 = (char *) malloc(len);
                                memset(h2, 0, len);
                                sscanf(hostname, "%d.%d.%d", &a, &b, &c);
                                for(d = 1; d <= 255; d++)
                                {
                                        snprintf(h2, len, host, a, b, c, d);
                                        if(plugin_next(h2) != PLUGIN_OK)
                                                mthis->usage
                                                        ("failure scanning class C");
                                }
                                return PLUGIN_OK;
                                break;
                        case 32:
                                range[0] = 0x00;
                                if(plugin_next(host_name) != PLUGIN_OK)
                                        mthis->usage("Invalid hostname/ip");
                                return PLUGIN_OK;
                                break;
                        default:
                                mthis->usage
                                        ("masks must be either /0 (whole internet), /8 (class A)"
                                         " /16 (class B), /24 (class C), or /32 (single host)");
                }

        }
        else
        {
                //regular hostname
                struct hostent *he;
                struct sa tp;

                if(mthis->ipverify(hostname))
                {
                        hosts->ip = inet_addr(hostname);
                        mthis->memput(&(hosts->host), hostname,
                                      strlen(hostname));
                }
                else
                {
                        he = gethostbyname(hostname);
                        if(he)
                        {
                                memcpy(&tp.addr, he->h_addr, he->h_length);
                                hosts->ip = tp.addr;
                                he = gethostbyaddr(he->h_addr, 4, AF_INET);
                                if(he)
                                        mthis->memput(&(hosts->host),
                                                      he->h_name,
                                                      strlen(he->h_name));
                                else
                                        mthis->memput(&(hosts->host),
                                                      hostname,
                                                      strlen(hostname));
                        }
                        else
                                return PLUGIN_ERROR;
                }

                if(hosts->ip == INADDR_NONE)
                {
                        free(hosts->host);
                        hosts->ip = 0;
                        return PLUGIN_ERROR;
                }

                hosts->next = (host_list_t *) malloc(sizeof(host_list_t));
                if(!hosts->next)
                {
                        my_errno = PLUGIN_ERROR;
                        return PLUGIN_ERROR;
                }
                hosts = hosts->next;
                hosts->next = NULL;
                if(!pthis.ready)
                        plugin_open(NULL, 0);
                num_hosts++;
                pthis.ready = true;
        }
        return PLUGIN_OK;
}

static void plugin_usage(void)
{
        assert(mthis != NULL);
        assert(mthis->output != NULL);
        mthis->output("\n ACK Scan plugin: requires root priveleges\n");
        mthis->output
                ("  Usage: [-pd sec] [-pt ttl] [-pp port] hostname/ip...\n");
        mthis->output("\t  -pd: delay for sec seconds after scans\n");
        mthis->output("\t  -pt: set IP TTL to ttl\n");
        mthis->output("\t  -pp: set destination port\n");
        mthis->output("  Prepend all hostnames for this plugin with 'A'\n");
        mthis->output("    Example: sonar Awww.google.com\n");
}

static void plugin_status(void)
{
        assert(mthis != NULL);
        assert(mthis->output != NULL);
        if(pthis.ready)
        {
                mthis->output("  ACK Scan plugin ready::\n");
                mthis->output("  TTL %d :: delay %d :: ", ttl, delay);
                if(port == -1)
                        mthis->output("random port\n");
                else
                        mthis->output("port %d\n", port);

        }
}

static int plugin_last_error(void)
{
        return my_errno;
}

static int rfc793_sweep(void *not_used)
{
        unsigned int counter = 0;
        //send an ack to port 0
        //if you get a RST back, host is up
        //if you timeout, host is down.
        // just make sure we, and sonar, are ready for processing

        if(!pthis.ready)
        {
                mthis->output("plugin not ready\n");
                return PLUGIN_ERROR;
        }
        assert(hosts != NULL);
        assert(first_host != NULL);
        hosts->next = first_host;
        hosts = first_host;

        while(hosts != NULL && counter != num_hosts)
        {
                if(hosts->ip == 0)
                {
                        //this is the "header"
                        hosts = hosts->next;
                        continue;
                }
                if(rfc793_send(hosts) == PLUGIN_ERROR)
                {
                        return PLUGIN_ERROR;
                }
                if(rfc793_recv(hosts) == PLUGIN_NORESP)
                {
                        struct in_addr s;
                        s.s_addr = hosts->ip;
                        mthis->message("no response from: %s[%s]\n",
                                       hosts->host, inet_ntoa(s));
                }
                else
                {

                }
                hosts = hosts->next;
                counter++;
        }
        sleep(delay);
        return PLUGIN_OK;
}

static int rfc793_send(host_list_t * hosts)
{
        unsigned char *packet;
	struct hostent *he;
	struct net_tuple nt;
	char name[256];
        int bwrote;
        int total_size = sizeof(struct tcphdr);

        {
                assert(hosts != NULL);
                assert(hosts->host != NULL);
                assert(hosts->ip != 0);
                assert(hosts->ip != (unsigned int) -1);
                assert(my_socket != 0);
                assert(my_socket != -1);
                assert(mthis->output != NULL);
        }

        srand(getpid() ^ (uid_t) time(NULL));
        packet = (unsigned char *) malloc(total_size);
        if(!packet)
        {
                mthis->output("could not allocate space for packet\n");
                return PLUGIN_ERROR;
        }
        memset(packet, 0, total_size);
	memset(name, 0, 256);

	gethostname(name, 255);
	he = gethostbyname(name);
        if(he)
        {
                memcpy(&nt.src, he->h_addr, he->h_length);;
        }
        else
                return PLUGIN_ERROR;	

	nt.dst = hosts->ip;

        nt.sport = htons(1 + (int) (65535.0 * rand() / (RAND_MAX + 1.0)));   //make this random
        if(port != -1)
                nt.dport = htons(port);
        else
                nt.dport =
                        htons(1 +
                              (int) (65535.0 * rand() / (RAND_MAX + 1.0)));

	add_tcphdr(packet, &nt, TH_ACK);
        // packet is ready to send       

        //time stamp our packet, for round trip time
        gettimeofday(&(tp), &(tz));

        bwrote = send_ipv4(my_socket, packet, total_size, &nt);
        if(bwrote == -1)
        {
                mthis->message("send_ipv4 failed: bailing...\n");
                perror("send_ipv4");
                return PLUGIN_ERROR;
        }
        if(bwrote < (int) total_size)
        {
                mthis->message("wrote %d bytes\n", bwrote);
        }
        free(packet);

        return PLUGIN_OK;
}

static int rfc793_recv(host_list_t * hosts)
{
        fd_set rfds;
        struct ip *iphdr;
        struct icmp *icmphdr;
        struct tcphdr *tcp;
        unsigned short iphdrlen;
        unsigned long triptime;
        struct timeval tv;
        int bread, ret;
        char buf[MAX_PACKET];
        int fromlen;
        struct sockaddr_in from;

        assert(my_socket != -1 && my_socket != 0);
        assert(err_socket != -1 && err_socket != 0);

        if(delay == 0)
                tv.tv_sec = 5;  // 5 seconds by default...
        else
                tv.tv_sec = delay;
        tv.tv_usec = 0;
      loop:
        memset(buf, 0, MAX_PACKET);
        FD_ZERO(&rfds);
        FD_SET(my_socket, &rfds);
        FD_SET(err_socket, &rfds);
        ret = select(((my_socket <
                       err_socket) ? my_socket + 1 : err_socket + 1), &rfds,
                     NULL, NULL, &tv);
        if(ret == -1)
        {
                mthis->message("could not select socket\n");
                return PLUGIN_ERROR;
        }
        if(ret == 0)
        {
                return PLUGIN_NORESP;
        }
        if(FD_ISSET(my_socket, &rfds))
        {
                bread = recvfrom(my_socket, buf, MAX_PACKET, 0,
                                 (struct sockaddr *) &(from),
                                 (socklen_t *) & (fromlen));
                if(bread == -1)
                {
                        perror("recvfrom");
                        mthis->message("recvfrom error\n");
                        return PLUGIN_ERROR;
                }

                gettimeofday(&tv, &(tz));

                tvsub(&tv, &tp);
                triptime = tv.tv_sec * 1000 + (tv.tv_usec / 1000);

                iphdr = (struct ip *) buf;
                iphdrlen = iphdr->ip_hl * 4;    // number of 32-bit words *4 = bytes

                if(bread < iphdrlen + sizeof(struct tcphdr))
                        mthis->message("Too few bytes from %s\n",
                                       inet_ntoa(from.sin_addr));

                tcp = (struct tcphdr *) (buf + iphdrlen);

                // if scanning round robin DNS hosts, this check may cause invalid reponses

                if(from.sin_addr.s_addr != hosts->ip)
                {
                        // we might get some other packets, so just ignore them...
                        goto loop;
                }

                if(!(tcp->th_flags & TH_RST))
                {
                        mthis->message
                                ("%s[%s]:%d online & filtered: range %lu ms ttl: %d size: %d\n",
                                 hosts->host, inet_ntoa(from.sin_addr),
                                 ntohs(tcp->th_sport), triptime, iphdr->ip_ttl,
                                 ntohs(iphdr->ip_len));
                        print_packet(tcp);
                        return PLUGIN_OK;
                }
                mthis->message
                        ("%s[%s]:%d online & unfiltered: range: %lu ms ttl: %d size: %d\n",
                         hosts->host, inet_ntoa(from.sin_addr),
                         ntohs(tcp->th_sport), triptime, iphdr->ip_ttl,
                         ntohs(iphdr->ip_len));
                print_packet(tcp);
                return PLUGIN_RESP;
        }
        else if(FD_ISSET(err_socket, &rfds))
        {
                bread = recvfrom(err_socket, buf, MAX_PACKET, 0,
                                 (struct sockaddr *) &(from),
                                 (socklen_t *) & (fromlen));
                if(bread == -1)
                {
                        perror("recvfrom");
                        mthis->message("recvfrom error\n");
                        return PLUGIN_ERROR;
                }

                gettimeofday(&tv, &(tz));

                tvsub(&tv, &tp);
                triptime = tv.tv_sec * 1000 + (tv.tv_usec / 1000);

                iphdr = (struct ip *) buf;
                iphdrlen = iphdr->ip_hl * 4;    // number of 32-bit words *4 = bytes

                if(bread < iphdrlen + sizeof(struct icmp))
                        mthis->message("Too few bytes from %s\n",
                                       inet_ntoa(from.sin_addr));

                icmphdr = (struct icmp *) (buf + iphdrlen);
                if(from.sin_addr.s_addr != hosts->ip)
                {
                        // we might get some other packets, so just ignore them...
                        goto loop;
                }

                if(icmphdr->icmp_type != ICMP_DEST_UNREACH && icmphdr->icmp_code != ICMP_UNREACH_PORT)
                {
                        mthis->message
                                ("bad ICMP response from host, abnormal behavior\n");
                        return PLUGIN_OK;
                }

                mthis->message
                        ("%s[%s]:%d online & filtered: range: %lu ms ttl: %d size: %d\n",
                         hosts->host, inet_ntoa(from.sin_addr),
                         ntohs(tcp->th_sport), triptime, iphdr->ip_ttl,
                         ntohs(iphdr->ip_len));

                return PLUGIN_RESP;
        }
        return PLUGIN_NORESP;
}

/**
  * @fn tvsub(register struct timeval *out, register struct timeval *in)
  * Subtract one struct timeval from another
  * @param out Pointer to destination (will be filled with out - in)
  * @param in Pointer to src
*/
static void tvsub(struct timeval *out, struct timeval *in)
{
        if((out->tv_usec -= in->tv_usec) < 0)
        {
                out->tv_sec--;
                out->tv_usec += 1000000;
        }
        out->tv_sec -= in->tv_sec;
}

static void print_packet(struct tcphdr *tcp)
{
        mthis->output("  sport: %d dport: %d\n", ntohs(tcp->th_sport),
                      ntohs(tcp->th_dport));
        mthis->output("  flags: (");
	if(tcp->th_flags & TH_FIN)
		mthis->output("F");
	else
		mthis->output("-");
	if(tcp->th_flags & TH_SYN)
		mthis->output("S");
	else
		mthis->output("-");
	if(tcp->th_flags & TH_RST)
		mthis->output("R");
	else
		mthis->output("-");
	if(tcp->th_flags & TH_PUSH)
		mthis->output("P");
	else
		mthis->output("-");
	if(tcp->th_flags & TH_ACK)
		mthis->output("A");
	else
		mthis->output("-");
	if(tcp->th_flags & TH_URG)
		mthis->output("U");
	else
		mthis->output("-");
        mthis->output(") offset: %d window: %d seq: %d ack: %d\n",
                      tcp->th_off, tcp->th_win, tcp->th_seq, tcp->th_ack);
}

int in_cksum(unsigned short *addr, int len)
{ 
   int nleft = len; 
   int sum = 0; 
   u_short *w = addr; 
   u_short answer = 0; 


   while (nleft > 1) 
   { 
      sum += *w++; 
      nleft -= 2; 
   } 


   if (nleft == 1) /* padding */ 
   { 
      *(u_char *) (&answer) = *(u_char *) w; 
      sum += answer; 
   } 


  return(sum); 
} 


void add_tcphdr(unsigned char *pkt, struct net_tuple *nt, uint8_t flags) 
{ 
   struct tcphdr tcp; 
   struct _fakehead fakehead; 
   int sum; 


   memset(&tcp, 0, sizeof(tcp)); 
   memset(&fakehead, 0, sizeof(fakehead)); 


   tcp.th_dport = nt->dport; 
   tcp.th_sport = nt->sport; 
   fakehead.saddr = nt->src; 
   fakehead.daddr = nt->dst; 
   fakehead.zero = 0, 
   fakehead.protocol = 6; 
   fakehead.tot_len = htons(TCP_SIZE); 
   sum = in_cksum((u_short *)&fakehead, sizeof(fakehead)); 
   tcp.th_off = TCP_SIZE >> 2; 
   tcp.th_seq = 1 + (int) (65535.0 * rand() / (RAND_MAX + 1.0)); 
   tcp.th_ack = 1 + (int) (65535.0 * rand() / (RAND_MAX + 1.0));
   tcp.th_flags |= flags; /* ADD the flags */ 
   tcp.th_win = 0; 
   sum += in_cksum((u_short *)&tcp, sizeof(tcp)); 
   tcp.th_sum = CKSUM_CARRY(sum); 
   memcpy(pkt, &tcp, sizeof(tcp)); 
} 






/* 
 * add's ipv4-header of 20 bytes without any options 
 * - IPPROTO_TCP and 40 bytes total length (NOT NEEDED YET)
 */ /*
void add_iphdr(unsigned char *pkt, struct net_tuple *nt) 
{ 
  struct ip ip; 


  memset(&ip, 0, sizeof(struct ip)); 
  ip.ip_hl = sizeof(ip) >> 2; 
  ip.ip_v = 4; 
  //ip->tos = 0;
  ip.ip_len = htons(IP_SIZE + TCP_SIZE);
  //ip->id = 0; done by kernel 
  //ip->frag_off = 0; 
  ip.ip_ttl = 0xff; 
  ip.ip_p = IPPROTO_TCP; 
  //ip->check = 0; done by kernel
  ip.ip_src.s_addr = nt->src; 
  ip.ip_dst.s_addr = nt->dst; 
  memcpy(pkt, &ip, sizeof(ip)); 
} */




/* 
 * send out ipv4-packet 
 * with data 'pkt' of length 'len' 
 * returns the number of characters sent, or -1 if an error occured 
 */ 
int send_ipv4(int sox, u_char *pkt, size_t len, struct net_tuple *nt) 
{ 
   struct sockaddr_in to; 


   to.sin_family = AF_INET; 
   memcpy(&to.sin_addr.s_addr, &nt->dst, sizeof(u_long)); 


   return(sendto(sox, pkt, len, 0 , (struct sockaddr *)&to, sizeof(to)) ); 
} 
