/***************************************************************************
                          sonar.c  -  main sonar file
                             -------------------
    begin                : Mon Apr  8 17:26:46 PDT 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: sonar.c,v 1.2 2002/08/08 04:06:00 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file sonar.c
  * File contains most of the main functions of sonar
  */
#include <assert.h>
#include "types.h"
#include "sonar.h"
#include "util.h"
#include "plugin.h"

static const char mainid[] =
        "\100$ autosec/sonar: $Id: sonar.c,v 1.2 2002/08/08 04:06:00 red0x Exp $";

//these are contained in plugin.cpp
extern plugin_list_t input_list[];
extern plugin_list_t output_list[];
extern plugin_list_t network_list[];
extern int num_inp, num_out, num_net;

static int count;
static int delay;

/**
  * @fn main(int argc, char *argv[])
  * Main Method
  * @param argc Number of arguments
  * @param argv Arguments
  * @return SONAR_OK on success
  */
int main(int argc, char *argv[])
{
        int cur = 0;
        int opt = 0;
	sonar_t mthis;
	raw_output = stderr;
        if(mainid)
        {
        }                       // useless reference to avoid warnings

        

        //setup mthis (sonar_t)
        mthis.ver_major = 0;
        mthis.ver_minor = 6;
        mthis.revision = 0;
        mthis.api = 2;
        mthis.resolver = &sonar_resolve;
        mthis.memput = &bput;
        mthis.message = &sonar_msg;
        mthis.ipverify = &sonar_isip;
        mthis.output = &sonar_rawoutput;
        mthis.usage = &sonar_usage;
        mthis.getprivs = &sonar_getprivs;
        mthis.dropprivs = &sonar_dropprivs;
        count = -1;
        delay = 300;

        //these will be fatal errors, as plugins we become the 
        // core of sonar
        if(plugin_start(&mthis) != PLUGIN_OK)
                sonar_usage("Could not initialize plugin API\n");
        if(plugin_loadall(&mthis) != PLUGIN_OK)
                sonar_usage("Could not load any plugins\n");

        while((opt = getopt(argc, argv, "fd:o:c:p:")) != -1)
        {
                // core options, handled first
                switch (opt)
                {
                        case 'f':
                                sfork++;
                                break;
                        case 'o':
                                sonar_setoutput(optarg);
                                break;
                        case 'c':
                                count = atoi(optarg);
                                if(count == 0)
                                        count = -1;
                                break;
                        case 'd':
                                delay = atoi(optarg);
                                if(delay < 1)
                                        delay = 1;
                                break;
                        case 'p':
                                if(plugin_command_handler
                                   (optarg[0], &optarg[1]) == SONAR_ERROR)
                                        sonar_usage("Argument not handled by plugin(s)");
                                break;
                        default:
                                sonar_usage("Bad Argument");
                                break;
                }
        }
        if(argv[optind] == NULL)
        {
                sonar_usage("no host/ip");
        }

        sonar_commandline(argv, argc, &cline);
        sonar_msg("%s\n", cline);

        while(argv[optind] != NULL && optind < argc)
        {
                // pass anything left off to network handlers
                if(plugin_network_command_handler(argv[optind]) ==
                   SONAR_ERROR)
                {
                        sonar_msg("not all hostnames/ips handled...\n");
                        return SONAR_ERROR;
                }
                optind++;
        }

        if(sfork)
        {
                sonar_msg("going into daemon mode, pid: %d\n", getpid());
                if(fork())
                        return SONAR_OK;

        }

        //signal handlers
        signal(SIGTERM, sig_term);
        signal(SIGINT, sig_int);
        signal(SIGKILL, sig_kill);
        signal(SIGHUP, sig_hup);

        /* Status report here */

        sonar_status_report();
        while(count == -1 || cur < count)
        {
                int err;
                if((err = sonar_sweep()) == PLUGIN_ERROR)
                {
                        sonar_msg("Sweep ended with a plugin error\n");
                        return SONAR_ERROR;
                }
                if(err == PLUGIN_OK)    // host online, and scanned ok, so...
                {

                }
                cur++;
		if(count > 1 || count == -1)
			sleep(delay);
        }

        if(plugin_killall() != PLUGIN_OK)
                sonar_msg("Could not unload plugins\n");
        if(plugin_end() != PLUGIN_OK)
                sonar_msg("could not destroy plugins\n");
        return SONAR_OK;
}

/** @fn sonar_getprivs(void)
  * Get root priveleges when needed
  * @return uid of user before we got root
  */
int sonar_getprivs(void)
{
        uid_t uid;
        uid = getuid();
        if(geteuid())
                sonar_usage("must be root!");
        return uid;
}

/** @fn sonar_dropprivs(uid_t uid)
  * Drop root priveleges when no longer needed
  * Drops down to specified uid
  * @param uid UID_T of user to drop to
  * @return SONAR_OK 
  */
int sonar_dropprivs(uid_t uid)
{
        if(uid != geteuid())    /* no need to drop priveleges, its not setuid */
        {
                sonar_msg("dropping privelegs to uid %d\n", uid);
                if(setuid(uid) != 0)
                {
                        sonar_msg("could not drop priveleges, halting\n");
                        exit(-1);
                }
        }
        return SONAR_OK;
}

/** @fn sonar_rawoutput(char *msg,...)
  * Output formatted text given by msg and any other arguments.
  * Used in the Plugin API
  * @param msg Format string to output (see printf)
  * @return SONAR_OK 
  */
int sonar_rawoutput(char *msg, ...)
{
        va_list ap;
        va_start(ap, msg);
        vfprintf(raw_output, msg, ap);
        va_end(ap);
        fflush(raw_output);
        return SONAR_OK;
}

/** @fn sonar_status_report()
  * Output a status report before each scan
  */
void sonar_status_report(void)
{
        sonar_rawoutput("sonar: ");
        sonar_rawoutput(" -delay %d-", delay);
        if(sfork)
                sonar_rawoutput(" -daemon mode-");
        if(exe)
                sonar_rawoutput(" -executing \"%s\"-", exe);
        if(count != -1)
                sonar_rawoutput(" -count %d-", count);
        sonar_rawoutput("\nPlugin report:\n");

        plugin_status_hook();
        sonar_rawoutput("\n");
}

/** @fn sonar_sweep()
  * Run one pass of a sonar sweep
  * @return SONAR_ERROR if sweep failed
  * @return SONAR_OK if sweep worked
  */
int sonar_sweep(void)
{
        //pick which network plugin to use, by getting the
        //ready one(s) and running their sweep functions.
        // these functions wont return until someone presses control-C, 
        // unless otherwise noted, at which point, program will return
        int status = PLUGIN_OK, plug;
        for(plug = 0; plug < num_net; plug++)
        {
                if(network_list[plug].plug == NULL)
                {
                        //broken plugin
                        DEBUG("BROKEN PLUGIN DETECTED\n");
                        continue;
                }
                {
                        assert(network_list[plug].plug != NULL);
                        assert(network_list[plug].plug->plugin_sweep != NULL);
                }
                // only run this sweep if the plugin is ready
                if(network_list[plug].plug->ready == true)
                        status = network_list[plug].plug->plugin_sweep(NULL);
                if(status != PLUGIN_OK)
                        return SONAR_ERROR;
        }
        return status;
}
