/***************************************************************************
                          util.cpp  -  description
                             -------------------
    begin                : Wed Apr 17 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: util.c,v 1.3 2002/08/22 21:57:13 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file util.c
  * File contains some helper functions
  */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "util.h"
#include "plugin.h"

extern int sonar_rawoutput(char *msg, ...);
extern int sonar_sweep(void);

static const char utilid[] =
        "\100$ autosec/sonar: $Id: util.c,v 1.3 2002/08/22 21:57:13 red0x Exp $";

/**
  * @fn sonar_usage(const char *err)
  * Print the usage
  * @param err Error to report
  */
void sonar_usage(const char *err)
{
        if(utilid)
        {
        }                       // useless reference to avoid warnings
        printf("%s %s - red0x <red0x@users.sourceforge.net>\n", PACKAGE,
               VERSION);
        printf("FREE for NON-COMMERCIAL USE\n");
        printf("sonar core functions:\n [-f] [-o output] [-d sec] [-c count] [plugin-opts]\n");
        printf("\n");
        printf("\t -f: fork into background (daemon mode)\n");
        printf("\t -o output: specify output file, defaults to stdout\n");
        printf("\t -d sec:  delay sec seconds between scans\n");
        printf("\t -c count: sweep count times (default: inifinite)\n");
        printf("\t plugin-opts: pass options to plugin\n\n");

        plugin_usage_hook();
	printf("\nFor bug reporting, please see:\n"
	       "<URL:http://autosec.sourceforge.net/helpDesk/>.\n");
        printf("\nERROR: %s\n", err);
        exit(-1);
}

/**
  * @fn bput(char ** dest, const char * src, const int len)
  * Copy memory from one place to another
  * @param dest Double pointer to destination
  * @param src Pointer the source
  * @param len Length of source
  */
void bput(char **dest, const char *src, const int len)
{
        *dest = (char *) malloc(len + 2);
        memset(*dest, 0, len + 2);
        strncpy(*dest, src, len);
//  return dest;
}

/**
  * @fn sonar_commandline(char *argv[], int argc, char **out)
  * Get the commandline
  * @param argv argv from main
  * @param argc argc from main
  * @param out Double pointer to destination
  */
void sonar_commandline(char *argv[], int argc, char **out)
{
        int i, len = 0;

        for(i = 0; i < argc; i++)
                len += strlen(argv[i]) + 1;

        *out = (char *) malloc(len + 2);
        memset(*out, 0, len + 2);

        for(i = 0; i < argc; i++)
        {
                strncat(*out, argv[i], strlen(argv[i]));
                strcat(*out, " ");
        }
}

/**
  * @fn sonar_msg(char *msg, ...)
  * Print a message with date and time
  * @param msg Format specifier like printf
  */
void sonar_msg(char *msg, ...)
{
        struct tm *gmt;
        time_t ltime;
        char *mtime;
        va_list ap;

        time(&ltime);
        gmt = localtime(&ltime);
        mtime = asctime(gmt);
        mtime[strlen(mtime) - 1] = '\0';
        sonar_rawoutput("[%s]: ", mtime);

        va_start(ap, msg);
        vfprintf(raw_output, msg, ap);
        va_end(ap);
        fflush(raw_output);
}

RETSIGTYPE sig_segv(int i)
{
        sonar_msg("Segmentation fault.\n");
	plugin_killall();
        exit(-2);
}

RETSIGTYPE sig_hup(int i)
{
        sonar_msg("caught SIGHUP, restarting..\n");
        sonar_sweep();
}

RETSIGTYPE sig_term(int i)
{
        sonar_msg("caught SIGTERM, cleaning up..\n");
        plugin_killall();
        fclose(raw_output);
        exit(0);
}

RETSIGTYPE sig_int(int i)
{
        sonar_msg("caught SIGINT, cleaning up..\n");
        plugin_killall();
        fclose(raw_output);
        exit(0);
}

RETSIGTYPE sig_kill(int i)
{
        sonar_msg("caught SIGKILL, cleaning up..\n");
	plugin_killall();
        /* if(session.rdy == 0)
           exit(0); */
        fclose(raw_output);
        exit(0);
}

/**
  * @fn sonar_setoutput(char *fname)
  * Set the output file
  * @param fname File name to output to
  */
int sonar_setoutput(char *fname)
{
        FILE *fp = fopen(fname, "a");
        if(fp)
        {
                raw_output = fp;
                return 0;
        }
        return -1;
}

/**
  * @fn sonar_isip(char *ip)
  * Check for a valid IP address
  * @param ip String representation of an IP address to check
  * @return 1 if the string is an IP, 0 otherwise.
  */
int sonar_isip(char *ip)
{
        int a, b, c, d;
        sscanf(ip, "%d.%d.%d.%d", &a, &b, &c, &d);
        if(a < 0)
                return 0;
        if(a > 255)
                return 0;
        if(b < 0)
                return 0;
        if(b > 255)
                return 0;
        if(c < 0)
                return 0;
        if(c > 255)
                return 0;
        if(d < 0)
                return 0;
        if(d > 255)
                return 0;
        return 1;
}

/**
  * @fn sonar_resolve(char *host)
  * Resolve a hostname/IP address in string format
  * @param host Hostname/ip in string format
  * @return unsigned long IP Address
  */
unsigned long sonar_resolve(char *host)
{
        struct hostent *he;
        struct sa tmp;
        if(sonar_isip(host))
                return (inet_addr(host));
        he = gethostbyname(host);
        if(he)
        {
                memcpy(&tmp.addr, he->h_addr, he->h_length);
        }
        else
                return (INADDR_NONE);
        return (tmp.addr);
}
